import os

from platformdirs import user_data_dir


def config_path() -> str | None:
    """
    Try to get the path of the config file from current directory or user_data_dir.

    Returns
    -------
    str | None
        The path of the config file.
    """

    appname = "paifulogger"
    appauthor = "Jim137"
    user_data_dir(appname, appauthor)
    if os.path.exists("./config.json"):
        return "."
    elif os.path.exists(f"{user_data_dir(appname, appauthor)}/config.json"):
        return user_data_dir(appname, appauthor)
    else:
        return None



import os
import sys

if sys.version_info < (3, 10):
    raise ImportError("Python 3.10 or above is required for PaifuLogger.")
del sys

main_path = os.path.dirname(os.path.abspath(__file__))
del os

from .version import __version__

from .log import config_path, log, log_paifu, log_parser, remove_old_paifu
from .paifu_dl import paifu_dl
from .src.get_paifu import get_paifu, get_paifu_from_client_log, get_paifu_from_local
from .src.i18n import localized_str

__all__ = [
    "__version__",
    "config_path",
    "get_paifu",
    "get_paifu_from_client_log",
    "get_paifu_from_local",
    "localized_str",
    "log",
    "log_paifu",
    "log_parser",
    "main_path",
    "remove_old_paifu",
    "paifu_dl",
]



import os
import sys

sys.path.append(os.getcwd())

from paifulogger import __version__


def build_pyproject():
    if "dev" in __version__:
        print("Dev version, not building.")
        return None

    with open("requirements.txt", "r") as f:
        requirements = [f"{line.strip()}" for line in f.readlines()]

    pyproject = f"""[build-system]
requires      = ["setuptools>=61.0.0", "wheel"]
build-backend = "setuptools.build_meta"

[project]
name = "PaifuLogger"
version = "{__version__}"
description = "Logging tenhou paifu into excel or html file with some key information."
readme = "README.md"
authors = [{{ name = "Jim137", email = "jim@mail.jim137.eu.org" }}]
license = {{ file = "LICENSE" }}
classifiers = [
    "License :: OSI Approved :: MIT License",
    "Programming Language :: Python :: 3.10",
    "Programming Language :: Python :: 3.11",
    "Programming Language :: Python :: 3.12",
]
keywords = ["mahjong", "review", "logger", "paifu", "tenhou", "haifu", "paipu", "riichi-mahjong"]
dependencies = {requirements}
requires-python = ">=3.10"

[project.urls]
Homepage = "https://github.com/Jim137/Tenhou-Paifu-Logger"

[project.scripts]
paifu = "paifulogger.__main__:main"
log = "paifulogger.log:main"
plog = "paifulogger.log:main"
plogger = "paifulogger.log:main"
paifulog = "paifulogger.log:main"
pdl = "paifulogger.paifu_dl:main"

[tool.setuptools.packages.find]
exclude = [
    "Paifu",
    "牌譜",
]

[tool.setuptools.package-data]
"paifulogger.localizations" = [
    "*.json",
]""".replace(
        "'", '"'
    )

    with open("pyproject.toml", "w") as f:
        f.write(pyproject)


if __name__ == "__main__":
    print(__version__)
    build_pyproject()



import io
import re

import pandas as pd

from .i18n import LocalStr
from .Paifu import Paifu


class PaifuHtml:
    def __init__(self, html_str: str, paifu_str, local_lang: LocalStr):
        if html_str.split("</tbody>")[0] == "":
            self.logged = ""
            self.create_html(paifu_str, local_lang)
        else:
            self.logged = html_str.split("</tbody>")[0]
        self.new_log = ""
        self.end_table = ""
        if html_str.split("</body>")[0] == "":
            self.replay = ""
        elif len(html_str.split("</body>")) < 2:
            self.replay = ""
        else:
            self.replay = html_str.split("</body>")[1].split("</html>")[0]
        self.end = "</body></html>"
        pass

    def __repr__(self) -> str:
        """
        Return the html string.

        Returns
        -------
        str
            The html string.
        """

        return self.logged + self.new_log + self.end_table + self.replay + self.end

    def create_html(self, paifu_str, local_lang: LocalStr) -> None:
        """
        Create the html file.

        Parameters
        ----------
        paifu_str: str
            The paifu string.
        local_lang: LocalStr
            The localized string.
        """

        self.logged += f"""<!DOCTYPE html>
        <html lang={local_lang.lang}>
        <head>
            <meta charset="utf-8">
            <title>{paifu_str}</title>
            <style>
                table {{
                    border-collapse: collapse;
                }}
                table, th, td {{
                    border: 1px solid black;
                }}
                th, td {{
                    padding: 5px;
                }}
                th {{
                    text-align: left;
                }}
            </style>
            <script>
                if (window.localStorage) {{
                    var p = document.querySelector('#persisted-text');
                    if (localStorage.text == null) {{
                        localStorage.text = p.value;
                    }} else {{
                        p.value = localStorage.text;
                    }}
                    p.addEventListener('keyup', function(){{ localStorage.text = p.value; }}, false);
                }}
            </script>
        </head>
        <body>
            <table style="width:100%">
                <thead>
                    <tr>
                        <th>{local_lang.date}</th>
                        <th>{local_lang.plc}</th>
                        <th>{local_lang.paifu}</th>
                        <th>{local_lang.remark}</th>
                        <th>{local_lang.preR}</th>
                        <th>{local_lang.r_change}</th>
                        <th>{local_lang.round_num}</th>
                        <th>{local_lang.win}</th>
                        <th>{local_lang.deal_in}</th>
                    </tr>
                </thead>
                <tbody>
        """

    def log_into_table(self, paifu: Paifu) -> None:
        """
        Log the paifu data into the html table.
        
        Parameters
        ----------
        paifu: Paifu
            The paifu data.
        """

        self.new_log += f"""
                    <tr>
                        <td>{paifu.time}</td>
                        <td>{paifu.get_place(paifu.ban)}</td>
                        <td><a href="{paifu.url}">{paifu.url}</a></td>
                        <td><textarea id="persisted-text"></textarea></td>
                        <td>{paifu.r[paifu.ban]}</td>
                        <td>{paifu.rate_change:.03f}</td>
                        <td>{paifu.get_round_num()}</td>
                        <td>{paifu.get_win_num(paifu.ban)}</td>
                        <td>{paifu.get_deal_in_num(paifu.ban)}</td>
                    </tr>
        """

    def _retrieve(self, local_lang: LocalStr) -> tuple[int, float, float, float]:
        """
        Retrieve the data from the html table.

        Parameters
        ----------
        local_lang: LocalStr
            The localized string.

        Returns
        -------
        tuple[int, float, float, float]
            The number of logged data, the average place, the win rate, and the deal-in rate.
        """

        pseudo_html = (
            self.logged
            + self.new_log
            + """
                </tbody>
            </table>
        </body>
        </html>"""
        )
        wrapper = io.StringIO(pseudo_html)
        df = pd.read_html(wrapper)[0]
        logged_num = df.shape[0] - 1
        avg_plc = df[f"{local_lang.plc}"].mean()
        win_rate = df[f"{local_lang.win}"].sum() / df[f"{local_lang.round_num}"].sum()
        deal_in_rate = (
            df[f"{local_lang.deal_in}"].sum() / df[f"{local_lang.round_num}"].sum()
        )
        return logged_num, avg_plc, win_rate, deal_in_rate

    def end_of_table(self, local_lang: LocalStr) -> None:
        """
        The end of the table.

        Parameters
        ----------
        local_lang: LocalStr
            The localized string.
        """

        logged_num, avg_plc, win_rate, deal_in_rate = self._retrieve(local_lang)
        self.end_table += f"""
                </tbody>
            </table>
            <p>{local_lang.log_num} = {logged_num}</p>
            <p>{local_lang.avg_plc} = {avg_plc:.2}</p>
            <p>{local_lang.win_rate} = {win_rate:.3%}</p>
            <p>{local_lang.deal_in_rate} = {deal_in_rate:.3%}</p>
        """


def log_into_html(paifu: Paifu, local_lang: LocalStr, output: str):
    """
    Log the paifu data into html file.

    Parameters
    ----------
    paifu: Paifu
        The paifu data.
    local_lang: LocalStr
        The localized string.
    output: str
        The output directory.
    """

    if paifu.player_num == 3:
        paifu_str = local_lang.sanma + local_lang.paifu
    else:
        paifu_str = local_lang.yonma + local_lang.paifu
    path = f"{output}/{local_lang.paifu}/{paifu_str}.html"
    try:
        with open(path, "r", encoding="utf-8") as f:
            html_str = f.read()
        html_c = PaifuHtml(html_str, paifu_str, local_lang)
    except FileNotFoundError:
        html_c = PaifuHtml("", paifu_str, local_lang)
    html_c.log_into_table(paifu)
    html_c.end_of_table(local_lang)
    with open(path, "w", encoding="utf-8") as f:
        f.write(repr(html_c))
    print(
        "html: "
        + local_lang.hint_record1
        + re.findall(r"\d{10}gm-\w{4}-\w{4}-\w{8}&tw=\d", paifu.url)[0]
        + local_lang.hint_record2
    )
    return None



import csv
import os.path
import re

import pandas as pd

from .i18n import LocalStr
from .Paifu import Paifu


def log_into_csv(paifu: Paifu, local_lang: LocalStr, output: str):
    """
    Log the paifu data into csv file.

    Parameters
    ----------
    paifu: Paifu
        The paifu data.
    local_lang: LocalStr
        The localized string.
    output: str
        The output directory.
    """

    if paifu.player_num == 3:
        paifu_str = local_lang.sanma + local_lang.paifu
    else:
        paifu_str = local_lang.yonma + local_lang.paifu
    path = f"{output}/{local_lang.paifu}/{paifu_str}.csv"
    if os.path.isfile(path):
        csvfile = open(path, "a+", encoding="utf-8")
        writer = csv.writer(csvfile)
    else:
        # Create a new csv file
        csvfile = open(path, "w", encoding="utf-8")
        writer = csv.writer(csvfile)
        writer.writerow(
            [
                "# id",
                local_lang.date,
                local_lang.plc,
                local_lang.paifu,
                local_lang.preR,
                local_lang.r_change,
                local_lang.round_num,
                local_lang.win,
                local_lang.deal_in,
            ]
        )
    try:
        try:
            # Read the csv file
            df = pd.read_csv(path, index_col=0, encoding="utf-8")
        except pd.errors.EmptyDataError:
            # If the csv file is empty, create a new DataFrame
            df = pd.DataFrame(columns=["id", "date", "plc", "paifu", "preR", "r_change", "round_num", "win", "deal_in"])
        # Append new paifu data to the csv file
        writer.writerow(
            [
                df.shape[0],
                paifu.time,
                paifu.get_place(paifu.ban),
                paifu.url,
                float(paifu.r[paifu.ban]),
                paifu.rate_change,
                paifu.get_round_num(),
                paifu.get_win_num(paifu.ban),
                paifu.get_deal_in_num(paifu.ban),
            ]
        )
        print(
            "csv: "
            + local_lang.hint_record1
            + re.findall(r"\d{10}gm-\w{4}-\w{4}-\w{8}&tw=\d", paifu.url)[0]
            + local_lang.hint_record2
        )
    finally:
        csvfile.close()
    return None



class LocalStr:
    lang: str
    main_path: str

    # ==============================================
    # Additional fields defined in localization file
    excelE: int

    avg_plc: str
    date: str
    deal_in: str
    deal_in_rate: str
    hint_duplicate: str
    hint_input: str
    hint_record1: str
    hint_record2: str
    hint_tw: str
    hint_url: str
    log2mjai_import_error: str
    log_num: str
    paifu: str
    plc: str
    preR: str
    r_change: str
    remark: str
    round_num: str
    sanma: str
    sanma_mjai_error: str
    win: str
    win_rate: str
    yonma: str
    # ==============================================

    def __init__(self, lang: str = "en", main_path: str = "./") -> None: ...
    def load_data(self, data) -> None: ...
    def __setattr__(self, name, val) -> None: ...
    def __getattr__(self, name): ...

def localized_str(lang: str, main_path: str = ...) -> LocalStr: ...



import re
from typing import cast

import openpyxl as xl
from openpyxl.worksheet.worksheet import Worksheet

from .i18n import LocalStr
from .Paifu import Paifu


def log_into_xlsx(paifu: Paifu, local_lang: LocalStr, output: str):
    """
    Log the paifu data into xlsx file.

    Parameters
    ----------
    paifu: Paifu
        The paifu data.
    local_lang: LocalStr
        The localized string.
    output: str
        The output directory.
    """

    try:
        if paifu.player_num == 3:
            paifu_str = local_lang.sanma + local_lang.paifu
        else:
            paifu_str = local_lang.yonma + local_lang.paifu
        path = f"{output}/{local_lang.paifu}/{paifu_str}.xlsx"
        wb = xl.load_workbook(path)

        # Prevent MyPy warning
        sheet = cast(Worksheet, wb.active)
        # Remove final statistics
        sheet.delete_rows(sheet.max_row)
    except FileNotFoundError:
        # Create a new workbook
        wb = xl.Workbook()
        sheet = cast(Worksheet, wb.active)
        sheet.append(
            [
                local_lang.date,
                local_lang.plc,
                local_lang.paifu,
                local_lang.remark,
                local_lang.preR,
                local_lang.r_change,
                local_lang.round_num,
                local_lang.win,
                local_lang.deal_in,
            ]
        )
        # set column width
        sheet.column_dimensions["A"].width = 20
        sheet.column_dimensions["C"].width = 71
        sheet.column_dimensions["E"].width = local_lang.excelE
    # Append new paifu data to the sheet
    sheet.append(
        [
            paifu.time,
            paifu.get_place(paifu.ban),
            paifu.url,
            "",
            float(paifu.r[paifu.ban]),
            paifu.rate_change,
            paifu.get_round_num(),
            paifu.get_win_num(paifu.ban),
            paifu.get_deal_in_num(paifu.ban),
        ]
    )
    sheet["C" + str(sheet.max_row)].style = "Hyperlink"
    # Add final statistics
    sheet.append(
        [
            local_lang.avg_plc,
            f"=average(B2:B{sheet.max_row})",
            "",
            local_lang.win_rate,
            f"=sum(H2:H{sheet.max_row})/sum(G2:G{sheet.max_row})",
            "",
            local_lang.deal_in_rate,
            f"=sum(I2:I{sheet.max_row})/sum(G2:G{sheet.max_row})",
        ]
    )
    sheet["E" + str(sheet.max_row)].style = "Percent"
    sheet["H" + str(sheet.max_row)].style = "Percent"
    wb.save(path)
    print(
        "xlsx: "
        + local_lang.hint_record1
        + re.findall(r"\d{10}gm-\w{4}-\w{4}-\w{8}&tw=\d", paifu.url)[0]
        + local_lang.hint_record2
    )
    return None



import argparse
import json
import os
import re
import sys
import urllib.error

from . import __version__, main_path
from .log import _get_lang
from .src.config import config_path
from .src.get_paifu import get_paifu
from .src.i18n import LocalStr, localized_str

REG_URL = r"https?://tenhou\.net/\d/\?log=\d{10}gm-\w{4}-\w{4}-\w{8}&tw=\d"


def _get_urls(urls, local_lang: LocalStr = localized_str("en", main_path)) -> list[str]:
    """
    Get urls from input or args.url.

    If remake, get urls from url_log.h5
    Else if not given args.url, get urls from input.

    Parameters
    ----------
    urls: str | list[str] | None
        The url of the game log.
    local_lang: LocalStr
        The localized string.

    Returns
    -------
    list
        The list of urls.

    Note
    ----
        `re.findall(REG_URL, url)` will return a list of urls that match the regular expression.
    """

    check_urls = []

    if not urls:
        for url in re.findall(REG_URL, input(local_lang.hint_input)):
            check_urls.append(url)
    else:
        for url in urls:
            check_urls.extend(re.findall(REG_URL, url))
    return check_urls


def paifu_dl(
    urls: str | list[str] | None = None,
    *,
    local_lang: LocalStr = localized_str("en", main_path),
    output: str = "./",
    mjai: bool = False,
) -> int:
    """
    Download paifu from tenhou.net.

    Parameters
    ----------
    urls: str | list[str] | None
        The url of the game log.
    local_lang: LocalStr
        The localized string.
    output: str
        The output directory.
    mjai: bool
        Output MJAI format paifu.

    Returns
    -------
    int
        The return code.
    """

    if isinstance(urls, str):
        urls = [urls]

    check_urls = _get_urls(urls, local_lang)

    retcode = 0
    for url in check_urls:
        try:
            get_paifu(url, local_lang, output, mjai)
            print(f"paifu_dl: {url} has been downloaded.")
        except urllib.error.URLError:
            print(local_lang.hint_url, url)
            retcode = 1
        except OSError:
            print(local_lang.hint_url, url)
            retcode = 1
    return retcode


def pdl(args: argparse.Namespace) -> int:

    local_lang = _get_lang(args.lang)

    return paifu_dl(
        args.url,
        local_lang=local_lang,
        output=args.output,
        mjai=args.mjai,
    )


def pdl_parser(
    config_path: str | None = None, parser: argparse.ArgumentParser | None = None
) -> argparse.Namespace:
    """
    Parse the arguments from the command line.

    Parameters
    ----------
    config_path: str | None
        The path of the config file.
    parser: argparse.ArgumentParser | None
        The parser.

    Returns
    -------
    argparse.Namespace
        The arguments.
    """
    config = {}
    if config_path and os.path.exists(f"{config_path}/config.json"):
        with open(f"{config_path}/config.json", "r") as f:
            config = json.load(f)

    if parser is None:
        parser = argparse.ArgumentParser(description="Paifu Downloader")
    parser.add_argument("url", nargs="*", help="URL of the match.")
    parser.add_argument(
        "-l",
        "--lang",
        type=str,
        help="""Language of the program and output files. Default is English.
        Available languages: English(en), 繁體中文(zh_tw), 简体中文(zh), 日本語(ja).""",
        default=config.get("lang", None),
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        help="Output directory. Default is './'.",
        default=config.get("output", None),
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        help="""Show version of the program.
        If this is used, all other arguments will be ignored and the program will be closed.""",
        version=f"%(prog)s {__version__}",
    )
    parser.add_argument(
        "--mjai",
        action="store_true",
        help="Output MJAI format paifu.",
        default=config.get("mjai", False),
    )

    args = parser.parse_args()
    if "pdl" in args.url:
        args.url.remove("pdl")
    return args


def main(plog_parser: argparse.ArgumentParser | None = None):
    args = pdl_parser(config_path(), plog_parser)
    sys.exit(pdl(args))



from datetime import datetime
import xml.etree.ElementTree as ET

class Paifu:
    url: str
    name: str
    time: datetime
    ban: int
    root: ET.Element
    go_type: int
    owari: list[str]
    r: list[str]
    plc: int
    rate_change: list[str]
    go_str: str
    player_num: int
    rounds: list[list[ET.Element]]
    def __init__(self, url: str, root: ET.Element) -> None: ...
    def go_type_distinguish(self) -> None: ...
    def get_place(self, ban) -> int: ...
    def get_rate_change(self) -> float: ...
    def get_round_num(self) -> int: ...
    def get_deal_in_num(self, ban) -> int: ...
    def get_win_num(self, ban) -> int: ...



import argparse
import sys

from paifulogger.log import main as plog_main
from paifulogger.paifu_dl import main as pdl_main
from paifulogger.version import __version__


def main():
    parser = argparse.ArgumentParser(description="PaifuLogger CLI", prog="PaifuLogger")
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        help="""Show version of the program.
        If this is used, all other arguments will be ignored and the program will be closed.""",
        version=f"%(prog)s {__version__}",
    )
    subparsers = parser.add_subparsers(required=False, dest="command")

    plog_parser = subparsers.add_parser(
        "plog", help="Paifu Logger. See 'paifulogger plog -h' for more info."
    )
    pdl_parser = subparsers.add_parser(
        "pdl", help="Paifu Downloader. See 'paifulogger pdl -h' for more info."
    )

    options = sys.argv

    if "plog" in options:
        sys.exit(plog_main(plog_parser))
    elif "pdl" in options:
        sys.exit(pdl_main(pdl_parser))
    else:
        parser.parse_args()
        sys.exit(parser.print_help())


if __name__ == "__main__":
    main()



import re
import xml.etree.ElementTree as ET
from datetime import datetime


class Paifu:
    def __init__(self, url: str, root: ET.Element):
        self.url = url
        self.name = url.split("=")[1] + "=" + url.split("=")[2]
        self.time = datetime.strptime(re.findall(r"\d{10}", self.url)[0], "%Y%m%d%H")
        self.ban = int(url[-1])
        self.root = root

        if gtype := root[1].get("type"):
            self.go_type = int(gtype)
        if owari := root[-1].get("owari"):
            self.owari = owari.split(",")
        if rate := root[2].get("rate"):
            self.r = rate.split(",")

        self.go_type_distinguish()
        self._rounds()
        self.plc = self.get_place(self.ban)
        self.rate_change = self.get_rate_change()

    def go_type_distinguish(self):
        """
        Distinguish the type of the game. And set the go_str and player_num.

        Examples
        --------
            三鳳南喰赤速 (go_type = 127) -> go_str = '三南喰赤速', player_num = 3

        """

        self.go_str = ""
        if self.go_type & 1:
            # PVP
            pass

        if self.go_type & 16:
            self.go_str += "三"
            self.player_num = 3
        else:
            self.go_str += "四"
            self.player_num = 4

        if self.go_type & 128:
            # 上
            pass

        if self.go_type & 32:
            # 特 or 鳳
            pass

        if self.go_type & 8:
            self.go_str += "南"
        else:
            self.go_str += "東"

        if self.go_type & 4:
            self.go_str += "喰"

        if not self.go_type & 2:
            self.go_str += "赤"

        if self.go_type & 64:
            self.go_str += "速"

    def _rounds(self):
        """
        Get rounds of the game.
        """
        self.rounds = [[] for _ in range(self.get_round_num() + 1)]
        round_idx = -1
        for el in self.root:
            # Each element has tag: str, attrib: dict, text, tail attributes
            if el.tag == "INIT":
                round_idx += 1
            self.rounds[round_idx].append(el)

    def get_place(self, ban) -> int:
        """
        Return the placing and rate before match

        Parameters
        ----------
        ban: int
            The player's seat number.
        """
        o0, s0, o1, s1, o2, s2, o3, s3 = self.owari

        if self.player_num == 4:
            sp = [float(s0), float(s1), float(s2), float(s3)]
            placing = [1, 1, 1, 1]
            for i in range(4):
                for j in range(4):
                    if sp[i] < sp[j]:
                        placing[i] += 1
        else:
            sp = [float(s0), float(s1), float(s2)]
            placing = [1, 1, 1]
            for i in range(3):
                for j in range(3):
                    if sp[i] < sp[j]:
                        placing[i] += 1
        return placing[ban]

    def get_rate_change(self) -> float:
        """
        Return the rate change after match.

        Returns
        -------
        float
            The rate change after match.

        Note
        ----
        Since the rate change has a correction of number of played games.
        We assumed that player has played over 400 games, which the correction is fixed to 0.2.
        """

        if self.player_num == 4:
            dr_result = (30, 10, -10, -30)
            corr = (sum([float(r) for r in self.r]) / 4 - float(self.r[self.ban])) / 40
            return 0.2 * (dr_result[self.plc - 1] + corr)
        else:
            dr_result = (30, 0, -30, 0)
            corr = (sum([float(r) for r in self.r]) / 3 - float(self.r[self.ban])) / 40
            if self.plc == 4:
                assert False, "Sanma has no 4th place."
            return 0.2 * (dr_result[self.plc - 1] + corr)

    def get_round_num(self) -> int:
        """
        Return the total number of rounds.

        Returns
        -------
        int
            The total number of rounds
        """
        return len(self.root.findall("INIT"))

    def get_deal_in_num(self, ban) -> int:
        """
        Return number of deal-ins.

        Parameters
        ----------
        ban: int
            The player's seat number.

        Returns
        -------
        int
            The number of deal-in
        """
        agaris = self.root.findall("AGARI")
        count = 0
        for agari in agaris:
            if agari.get("fromWho") == str(ban) and agari.get("who") != str(ban):
                count += 1
        return count

    def get_win_num(self, ban) -> int:
        """
        Return number of wins.

        Parameters
        ----------
        ban: int
            The player's seat number.

        Returns
        -------
        int
            The number of wins
        """
        agaris = self.root.findall("AGARI")
        count = 0
        for agari in agaris:
            if agari.get("who") == str(ban):
                count += 1
        return count



import gzip
import os
import urllib.request
import xml.etree.ElementTree as ET
from glob import glob

from .i18n import LocalStr, localized_str
from .Paifu import Paifu

HEADER = {
    "Host": "e.mjv.jp",
    "User-Agent": "Mozilla/5.0 (X11; Linux x86_64; rv:65.0) Gecko/20100101 Firefox/65.0",
    "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8",
    "Accept-Language": "en-US,en;q=0.5",
    "Accept-Encoding": "gzip, deflate",
    "Connection": "keep-alive",
}


def url_request_handler(url: str) -> str:
    """
    Request the url and return the response from Tenhou server.

    Parameters
    ----------
    url: str
        The url of the game log.

    Returns
    -------
    str
        The response from Tenhou server.
    """
    url = url.split("=")[1]
    url = "https://tenhou.net/0/log/?" + url[:-3]
    req = urllib.request.Request(url=url, headers=HEADER)
    opener = urllib.request.build_opener()
    response = opener.open(req)
    response = gzip.decompress(response.read()).decode("utf-8")
    return response


def get_paifu(
    url: str,
    local_lang: LocalStr = localized_str("en"),
    output: str = "./",
    mjai: bool = False,
    no_output: bool = False,
) -> Paifu:
    """
    Get the paifu data from the given URL.

    Parameters
    ----------
    url: str
        The URL of the game log.
    local_lang: LocalStr
        The localized string.
    output: str
        The output directory.
    mjai: bool
        If True, output the paifu data in MJAI format.
    no_output: bool
        If True, do not output the paifu data.

    Returns
    -------
    Paifu
        The paifu data.
    """

    response = url_request_handler(url)
    root = ET.fromstring(response)
    paifu = Paifu(url, root)

    if no_output:
        return paifu

    path = f"{output}/{local_lang.paifu}/{paifu.go_str}/"
    if not os.path.isdir(path):
        os.makedirs(path)

    with open(path + paifu.name + ".xml", "w") as t:
        t.write(response)

    # mjai format output
    if mjai:
        save_mjai(paifu, path, paifu.name, local_lang)

    return paifu


def save_mjai(
    paifu: Paifu, path: str, url: str, local_lang: LocalStr = localized_str("en")
):
    """
    Save the paifu data in MJAI format.

    Parameters
    ----------
    paifu: Paifu
        The paifu data.
    path: str
        The output directory.
    url: str
        The URL of the game log.
    local_lang: LocalStr
        The localized string.

    Returns
    -------
    None
    """

    if paifu.player_num == 3:
        print(local_lang.sanma_mjai_error)
        return
    try:
        from .mjlog2mjai.parse import load_mjlog, parse_mjlog_to_mjai
    except ImportError:
        print(local_lang.log2mjai_import_error)
        return
    if not os.path.isdir(path + "/mjai/"):
        os.makedirs(path + "/mjai/")
    with open(path + "/mjai/" + url + ".mjson", "w", encoding="UTF-8") as f:
        f.write(parse_mjlog_to_mjai(load_mjlog(path + url + ".xml")))
    return


def get_paifu_from_local(
    url: str,
    go_str: str,
    local_lang: LocalStr = localized_str("en"),
    output: str = "./",
) -> Paifu | None:
    """
    Get the paifu data from the local file.

    Parameters
    ----------
    url: str
        The URL of the game log.
    go_str: str
        The go string of the game log.
    local_lang: LocalStr
        The localized string.
    output: str
        The output directory.

    Returns
    -------
    Paifu
        The paifu data.
    """

    path = f"{output}/{local_lang.paifu}/{go_str}/"
    url = url.split("=")[1] + "=" + url.split("=")[2]
    try:
        with open(path + url + ".xml", "r") as t:
            response = t.read()
    except FileNotFoundError:
        print(f"Cannot find {path + url + '.xml'}")
        return None
    root = ET.fromstring(response)
    paifu = Paifu(url, root)
    return paifu


def get_paifu_from_client_log(
    path: str,
) -> list[Paifu] | None:
    """
    Get the paifu data from the client log.

    Parameters
    ----------
    path: str
        The path of the client log.

    Returns
    -------
    list[Paifu]
        The list of paifu data.
    """

    if not os.path.isdir(path):
        print(f"Cannot find {path}")
        return None
    paifu_list = []
    for file in glob(f"{path}/*.mjlog"):
        url = "https://tenhou.net/0/?log=" + file.split("/")[-1][:-6]
        with gzip.open(file, "r") as f:
            file_content = f.read()
        root = ET.fromstring(file_content)
        paifu = Paifu(url, root)
        paifu_list.append(paifu)
    return paifu_list



import argparse
import json
import os
import re
import sys
import urllib.request
import warnings
from typing import Callable

from pandas import DataFrame, HDFStore

from . import __version__, main_path
from .src.config import config_path
from .src.get_paifu import get_paifu, get_paifu_from_client_log, save_mjai
from .src.i18n import LocalStr, localized_str
from .src.log_into_csv import log_into_csv
from .src.log_into_html import log_into_html
from .src.log_into_xlsx import log_into_xlsx
from .src.Paifu import Paifu
from .src.url_log import check_duplicate, url_log

REG_URL = r"https?://tenhou\.net/\d/\?log=\d{10}gm-\w{4}-\w{4}-\w{8}&tw=\d"
avaiable_formats = ["xlsx", "html", "csv"]


def remove_old_paifu(paifu_str: str, formats: list[str], output: str) -> None:
    """
    Remove old paifu files.

    Parameters
    ----------
    paifu_str : str
        The name of the paifu file.
    formats : list
        The formats of the paifu file.
    output : str
        The output directory.

    Returns
    -------
    None

    Examples
    --------
    >>> remove_old_paifu("paifu", ["xlsx", "html", "csv"], "./")
    paifu.xlsx has been removed.
    paifu.html has been removed.
    paifu.csv has been removed.
    """

    for format in formats:
        if os.path.exists(f"{output}/{paifu_str}.{format}"):
            os.remove(f"{output}/{paifu_str}.{format}")
            print(f"{paifu_str}.{format} has been removed.")
    return None


def _get_lang(lang: str | None = None) -> LocalStr:
    """
    Get the localized string.

    If not given, use English.

    Parameters
    ----------
    lang : str
        The language of the program.

    Returns
    -------
    LocalStr
        The localized string.
    """

    return localized_str(lang, main_path) if lang else localized_str("en", main_path)


def _get_output(output: str = "./") -> str:
    """
    Get output directory from args.output.

    If not given, use current directory "./".

    Parameters
    ----------
    output : str
        The output directory.

    Returns
    -------
    str
        The absolute path of the output directory.
    """

    return os.path.abspath(output) if output else os.path.abspath("./")


def _get_urls(
    url: list[str] | None = None,
    local_lang: LocalStr = localized_str("en"),
    output: str = os.path.abspath("./"),
    remake: bool = False,
) -> list[str]:
    """
    Get urls from input or url.

    If remake, get urls from url_log.h5
    Else if not given url, get urls from input.

    Parameters
    ----------
    url : list[str]
        The urls of the paifu files.
    local_lang : LocalStr
        The localized string.
    output : str
        The output directory.
    remake : bool
        Remake the log file from url_log.h5.

    Returns
    -------
    list[str]
        The list of urls.

    Note
    ----
    `re.findall(REG_URL, _url)` will return a list of urls that match the regular expression.
    """

    urls = []
    if remake:
        with HDFStore(f"{output}/{local_lang.paifu}/url_log.h5") as store:
            # ===============================================================
            # Special case: if "url" not in store, add it.
            # It will be deprecated in the future.
            if "url" not in store:
                store["url"] = DataFrame(columns=["url"])
                warnings.warn(
                    """The url_log.h5 you used is deprecated.
                    You have to manually copy all urls and delete url_log.h5,
                    then run and paste the urls to the program.
                    The new url_log.h5 will be automatically created.
                    """,
                    DeprecationWarning,
                )
            # ===============================================================
            urlstore = store["url"]["url"].values
            for _url in urlstore:
                if not re.match(REG_URL, _url):
                    urls.append("https://" + _url)
                    continue
                urls.append(_url)
    elif not url:
        for _url in re.findall(REG_URL, input(local_lang.hint_input)):
            urls.append(_url)
    else:
        for _url in url:
            urls.extend(re.findall(REG_URL, _url))
    return urls


def _get_formats(format: list[str] | None = None) -> list:
    """
    Get formats from `args.format`.

    If not given, return ["csv"].

    Parameters
    ----------
    format : list
        The formats of the output file.

    Returns
    -------
    list
        The formats of the output file.
    """

    return format if format else ["csv"]


def _remake_log(
    local_lang: LocalStr, output: str, formats: list[str], all_formats: bool = False
) -> None:
    """
    Remake the log file from url_log.h5 (past logging log), and remove old paifu files.

    Parameters
    ----------
    local_lang : LocalStr
        The localized string.
    output : str
        The output directory to remove.
    formats : list
        The formats of the output file.
    all_formats : bool
        Remove all formats.
    """

    paifu_str3 = local_lang.paifu + "/" + local_lang.sanma + local_lang.paifu
    paifu_str4 = local_lang.paifu + "/" + local_lang.yonma + local_lang.paifu
    try:
        if all_formats:
            remove_old_paifu(paifu_str3, avaiable_formats, output)
        else:
            remove_old_paifu(paifu_str3, formats, output)
    except OSError:
        pass
    try:
        if all_formats:
            remove_old_paifu(paifu_str4, avaiable_formats, output)
        else:
            remove_old_paifu(paifu_str4, formats, output)
    except OSError:
        pass
    return None


def _get_log_func(formats: list[str], all_formats: bool = False) -> list[Callable]:
    """
    Parse the formats and return the corresponding log functions.

    If `all_formats`, return all log functions.

    Parameters
    ----------
    formats : list
        The formats of the output file.
    all_formats : bool
        Output all formats.

    Returns
    -------
    list
        The list of log functions.
    """

    assert formats, "No format is given."

    if all_formats:
        log_formats = [
            log_into_xlsx,
            log_into_html,
            log_into_csv,
        ]
    else:
        log_formats = []
        if "xlsx" in formats:
            log_formats.append(log_into_xlsx)
        if "html" in formats:
            log_formats.append(log_into_html)
        if "csv" in formats:
            log_formats.append(log_into_csv)
    return log_formats


def log_paifu_from_local(
    paifus: list[Paifu],
    *,
    log_formats: list[Callable] = [log_into_csv],
    local_lang: LocalStr = localized_str("en", main_path),
    output: str = os.path.abspath("./"),
    ignore_duplicated: bool = False,
    mjai: bool = False,
):
    """
    Log paifu files from local.

    Parameters
    ----------
    paifus : list[Paifu]
        The list of paifu files.
    log_formats : list
        The log functions.
    local_lang : LocalStr
        The localized string.
    output : str
        The output directory.
    ignore_duplicated : bool
        Ignore duplicated urls.
    mjai : bool
        Output MJAI format paifu.

    Returns
    -------
    int
        The return code.
    """

    retcode = 0
    for paifu in paifus:
        try:
            if mjai:
                save_mjai(paifu, output, paifu.name, local_lang)
            if check_duplicate(paifu.url, local_lang, output) and not ignore_duplicated:
                print(local_lang.hint_duplicate, paifu.url)
                continue
            for log_into_format in log_formats:
                log_into_format(paifu, local_lang, output)
            url_log(paifu.url, local_lang, output)
        except OSError:
            print(local_lang.hint_url, paifu.url)
            retcode = 1
        except ValueError:
            print(local_lang.hint_tw, paifu.url)
            retcode = 1
        except KeyError:
            print(
                """Please remake the log file by `plog -f [format] -l [lang] -o [output] -r` first,
                  or `python -m paifulogger plog -f [format] -l [lang] -o [output] -r` manually."""
            )
            return 1
    return retcode


def log_paifu(
    *urls: list[str] | str,
    log_formats: list[Callable] = [log_into_csv],
    local_lang: LocalStr = localized_str("en", main_path),
    output: str = os.path.abspath("./"),
    remake: bool = False,
    ignore_duplicated: bool = False,
    mjai: bool = False,
) -> int:
    """
    Log paifu files.

    Parameters
    ----------
    urls : list
        The urls of the paifu files.
    log_formats : list
        The log functions.
    local_lang : LocalStr
        The localized string.
    output : str
        The output directory.
    remake : bool
        Remake the log file from url_log.h5.
    ignore_duplicated : bool
        Ignore duplicated urls.
    mjai : bool
        Output MJAI format paifu.

    Returns
    -------
    int
        The return code.
    """

    retcode = 0
    _urls: list[str] = []
    for url in urls:
        if isinstance(url, list):
            _urls.extend(url)
        else:
            _urls.append(url)
    for url in _urls:
        if not re.match(REG_URL, url):
            print(local_lang.hint_url, url)
            continue
        if remake:
            pass
        elif check_duplicate(url, local_lang, output) and not ignore_duplicated:
            print(local_lang.hint_duplicate, url)
            continue
        try:
            paifu = get_paifu(url, local_lang, output, mjai)
            for log_into_format in log_formats:
                log_into_format(paifu, local_lang, output)
            if remake:
                pass
            else:
                url_log(url, local_lang, output)
        except urllib.error.URLError:
            print(local_lang.hint_url, url)
            retcode = 1
        except OSError:
            print(local_lang.hint_url, url)
            retcode = 1
        except ValueError:
            print(local_lang.hint_tw, url)
            retcode = 1
        except KeyError:
            print(
                """Please remake the log file by `plog -f [format] -l [lang] -o [output] -r`,
                  or `python -m paifulogger plog -f [format] -l [lang] -o [output] -r` manually."""
            )
            return 1
    return retcode


def log(args: argparse.Namespace) -> int:
    """
    Parse the arguments and log paifu files.

    Parameters
    ----------
    args : argparse.Namespace
        The arguments from the command line.

    Returns
    -------
    int
        The return code.
    """

    local_lang = _get_lang(args.lang)
    output = _get_output(args.output)
    formats = _get_formats(args.format)

    # if remake, remove old files
    if args.remake:
        _remake_log(local_lang, output, formats, args.all_formats)

    log_formats = _get_log_func(formats, args.all_formats)

    if args.from_client:
        paifus = get_paifu_from_client_log(args.from_client)
        if not paifus:
            print("No paifu files found.")
            return 0
        print(f"Found {len(paifus)} paifu files.")
        return log_paifu_from_local(
            paifus,
            log_formats=log_formats,
            local_lang=local_lang,
            output=output,
            ignore_duplicated=args.ignore_duplicated,
            mjai=args.mjai,
        )

    urls = _get_urls(args.url, local_lang, output, args.remake)
    return log_paifu(
        urls,
        log_formats=log_formats,
        local_lang=local_lang,
        output=output,
        remake=args.remake,
        ignore_duplicated=args.ignore_duplicated,
        mjai=args.mjai,
    )


def log_parser(
    config_path: str | None = None, parser: argparse.ArgumentParser | None = None
) -> argparse.Namespace:
    """
    Parse the arguments from the command line.

    Parameters
    ----------
    config_path : str
        The path of the config file.
    parser : argparse.ArgumentParser
        The argument parser.

    Returns
    -------
    argparse.Namespace
        The arguments from the command line.
    """

    config = {}
    if config_path and os.path.exists(f"{config_path}/config.json"):
        with open(f"{config_path}/config.json", "r") as f:
            config = json.load(f)

    if parser is None:
        parser = argparse.ArgumentParser(description="Paifu Logger", prog="PaifuLogger")
    parser.add_argument("url", nargs="*", help="URL of the match.")
    parser.add_argument(
        "-l",
        "--lang",
        type=str,
        help="""Language of the program and output files. Default is English.
        Available languages: English(en), 繁體中文(zh_tw), 简体中文(zh), 日本語(ja).""",
        default=config.get("lang", None),
    )
    parser.add_argument(
        "-f",
        "--format",
        action="append",
        type=str,
        help="Format of the output file. Default is csv. Available formats: xlsx, html, csv.",
        choices=avaiable_formats,
        default=config.get("format", None),
    )
    parser.add_argument(
        "-a",
        "--all-formats",
        action="store_true",
        help="Output all formats.",
        default=config.get("all_formats", False),
    )
    parser.add_argument(
        "-r",
        "--remake",
        action="store_true",
        help="""Remake the log file from url_log.h5 (past logging log).
        Use this when the program is updated, changing format or language of the log file, or the log file is missing.
        Note that this will overwrite the log file.""",
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        help="Output directory. Default is './'.",
        default=config.get("output", None),
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        help="""Show version of the program.
        If this is used, all other arguments will be ignored and the program will be closed.""",
        version=f"%(prog)s {__version__}",
    )
    parser.add_argument(
        "--mjai",
        action="store_true",
        help="Output MJAI format paifu.",
        default=config.get("mjai", False),
    )
    parser.add_argument(
        "-c",
        "--from-client",
        nargs="+",
        help="Log client saved paifu (*.mjlog) from given directory.",
        type=str,
    )
    # Args for Debugging
    parser.add_argument(
        "--ignore-duplicated",
        action="store_true",
        help=argparse.SUPPRESS,
        default=config.get("ignore_duplicated", False),
    )
    args = parser.parse_args()
    if "plog" in args.url:
        args.url.remove("plog")
    return args


def main(plog_parser: argparse.ArgumentParser | None = None):
    args = log_parser(config_path(), plog_parser)
    return log(args)


if __name__ == "__main__":
    sys.exit(main())
